"""Garfield: Graph-based Contrastive Learning enable Fast Single-Cell Embedding"""
from ._settings import settings
from ._version import __version__

# import sys
# # needed when building doc (borrowed from scanpy)
# sys.modules.update(
#     {f'{__name__}.{m}': globals()[m] for m in ['pp']}) # ['tl', 'pp', 'pl']
    
